"""This file contains the Model class

Raoul Collenteur, 2025

"""

import matplotlib.pyplot as plt
from numpy import nan
from pandas import DataFrame, IntervalIndex, Series

from .core import validate_data


class Model:
    def __init__(self, wt, name=None, mcr=None):
        """Basic model to estimate groundwater recharge from water table data.

        Parameters
        ----------
        wt : pd.Series
            Time series of the water table fluctuations. The index should be a
            DatetimeIndex.
        name : str, optional
            Name of the model, by default None
        mcr : MCR instance, optional
            Instance of the MCR class to use for extrapolation of the water table
            before computing the rises, by default None

        Notes
        -----
        The Model class can be used to estimate groundwater recharge from water table
        data. The recharge is estimated as the product of the rises in the water table
        and the specific yield (sy). The rises can be computed using different methods,
        which can be specified using the rise_rule parameter in the estimate method.

        If an instance of the MCR class is provided, the Master Recession Curve (MCR)
        is used to extrapolate the water table before computing the rises.

        """
        validate_data(wt)
        self.wt = wt.dropna()
        self.name = name

        # Settings dict
        self.settings = {
            "freq": "D",
            "use_mcr": None,  # None means use MCR when available
        }

        self.parameters = DataFrame(
            index=["sy"],
            data=[[0.1, 0.1, 0.005]],
            columns=["initial", "optimal", "stderr"],
        )

        self.filter = None
        self.mcr = mcr
        self.events = None

    def fit_mcr(self, tmin=None, tmax=None):
        """Method to fit the Master Recession Curve (MCR).

        Parameters
        -----------
        tmin: pd.Timestamp
            Start of the period to use for the fitting. If None, the start of the time
            series is used.
        tmax: pd.Timestamp
            End of the period to use for the fitting. If None, the end of the time
            series is used.

        Notes
        -----


        """
        if self.mcr is None:
            raise ValueError(
                "No MCR instance provided. Please provide an instance of "
                "the MCR class when creating the Model instance."
            )

        if tmin is None:
            tmin = self.wt.index.min()
        if tmax is None:
            tmax = self.wt.index.max()

        wt_fit = self.wt[(self.wt.index >= tmin) & (self.wt.index <= tmax)]

        self.mcr.fit_mcr(wt_fit)

    def get_recharge_event_intervals(self, wt: Series, rise_rule="rises") -> Series:
        """Method to get the intervals on which to compute the recharge.

        Parameters
        ----------
        wt : pd.Series
            Time series of the water table fluctuations.

        Returns
        -------
        pd.IntervalIndex
            IntervalIndex with the events to compute the recharge on.

        """
        dt = wt.index.diff()[1:]
        dh = wt.diff()[1:]

        # Create an IntervalIndex for which the rises are computed
        dtint = IntervalIndex.from_arrays(left=dh.index - dt, right=dh.index)

        changes = Series(index=dtint, data=dh.values)

        # 1. Get the points for which to consider the rises
        if rise_rule not in ["both", "rises"]:
            raise ValueError("rise_rule should be 'both' or 'rises'.")

        # Get only the rise sections
        if rise_rule == "rises":
            events_int = changes[changes.values > 0].index
        # Get every point
        elif rise_rule == "both":
            events_int = changes.index

        if events_int is None or events_int.empty:
            raise ValueError(
                "No recharge events found. Please check the water table "
                "data and the rise_method."
            )
        self.events = events_int
        return events_int

    def estimate_recharge(self, sy=None, fit_mcr=False, rise_rule="rises") -> Series:
        """Method to estimate the groundwater recharge.

        Parameters
        ----------
        use_mcr: bool
            Set True to use the Master Recession Curve to extrapolate the watertable
            before computing the rises. Set False to not extrapolate. Defaults to None,
            which means that the MCR is used when present.

        Returns
        -------
        pandas.Series:
            Pandas Series with a DatetimeIndex and the values of the recharge estimates.

        Notes
        -----
        The recharge is estimated as the product of the rises and the specific yield.
        The recharge is resampled to the frequency defined in self.settings['freq'].
        Missing values are filled by backfilling.

        The following steps are performed:
        1. If fit_mcr is True, the MCR is fitted to the water table data.
        2. The rises are extracted from the water table data using the rise_method.
        3. The recharge is computed as the product of the rises and the specific yield.
        4. The recharge is resampled to the frequency defined in self.settings['freq'].
           Missing values are filled by backfilling.

        """
        # 1. Fit the MCR if requested
        if fit_mcr:
            self.fit_mcr()

        # 2. Get the points for which to consider the rises
        events_int = self.get_recharge_event_intervals(self.wt, rise_rule=rise_rule)

        # 3. Compute the rises
        if self.mcr is not None:
            # 3. Extrapolate using the MCR
            left_hand = self.mcr.get_extrapolated(self.wt, events_int)

        else:
            left_hand = Series(index=events_int.left, data=self.wt[events_int.left])

        #
        rises = Series(
            index=events_int,
            data=self.wt[events_int.right].values - left_hand.values,
        )

        rises = rises[rises.values > 0]
        self.rises = rises

        if sy is None:
            if self.parameters.loc["sy", "optimal"] is nan:
                raise ValueError(
                    "The specific yield is not set. Please provide a value "
                    "for sy or set the optimal value using the "
                    "self.parameters DataFrame."
                )
            else:
                sy = self.parameters.loc["sy", "optimal"]
        else:
            self.parameters.loc["sy", "optimal"] = sy

        # 3. Compute the recharge as the product of the events and the specific yield
        recharge = rises * sy

        # Set the index to the right side of the intervals
        recharge.index = rises.index.right
        recharge.name = "recharge"

        # Clip negative values to zero
        recharge[recharge < 0] = 0

        return recharge

    def plot(self, ax=None, figsize=(10, 6)):
        """Method to plot the water table, the rises and the recharge.

        Parameters
        ----------
        ax: matplotlib.axes.Axes or None
            Axes to plot the water table, rises and recharge on. If None, a new
            figure and axes are created.

        Returns
        -------
        ax: matplotlib.axes.Axes
            Axes with the water table, rises and recharge.


        """
        if ax is None:
            _, axs = plt.subplots(2, 1, sharex=True, figsize=figsize)

        # Plot the water table
        self.wt.plot(ax=axs[0], label="Water table", marker=".", ls="none", color="k")

        # Plot the rises
        if self.events is not None:
            for interval, rise in self.rises.items():
                axs[0].plot(
                    [interval.left, interval.right],
                    [self.wt.loc[interval.left], self.wt.loc[interval.left] + rise],
                    lw=3,
                    label="Rises",
                    color="C1",
                )

        # Plot the recharge
        if self.events is not None:
            self.estimate_recharge().plot(ax=axs[1], label="Recharge")

        axs[0].set_ylabel("Water table [m]")
        axs[0].legend(["Water table", "Rises"])

        axs[1].set_ylabel("Recharge [m/d]")
        axs[1].set_xlabel("Time")

        plt.tight_layout()

        return axs



import numpy as np
import pandas as pd
import pytest

from gwtf.mcr import MCR


@pytest.fixture
def simple_wt():
    idx = pd.date_range("2020-01-01", periods=10, freq="D")
    data = np.linspace(10, 20, 10)
    return pd.Series(data, index=idx)


def test_mcr_init():
    mcr = MCR()
    assert mcr.name == "MCR"
    assert mcr.nparam == 2


def test_get_fall_sections(simple_wt):
    mcr = MCR()
    falls = mcr.get_fall_sections(simple_wt)
    assert isinstance(falls, pd.DataFrame)


def test_get_dhdt(simple_wt):
    mcr = MCR()
    dhdt = mcr.get_dhdt(simple_wt)
    assert isinstance(dhdt, pd.Series)


def test_estimate_dhdt(simple_wt):
    mcr = MCR()
    # Set parameters for test
    mcr.parameters.loc[:, "optimal"] = [0.1, 0.2]
    dhdt_est = mcr.estimate_dhdt(simple_wt)
    assert isinstance(dhdt_est, pd.Series)


def test_fit_mcr_raises(simple_wt):
    mcr = MCR()
    # Test should raise an error
    with pytest.raises(Exception):
        mcr.fit_mcr(simple_wt)


def test_get_extrapolated(simple_wt):
    mcr = MCR()
    mcr.parameters.loc[:, "optimal"] = [0.1, 0.2]
    # Create dummy events
    idx = pd.date_range("2020-01-02", periods=5, freq="D")
    from pandas import IntervalIndex

    events = IntervalIndex.from_arrays(left=idx - pd.Timedelta(days=1), right=idx)
    result = mcr.get_extrapolated(simple_wt, events)
    assert isinstance(result, pd.Series)


def test_plot_runs(simple_wt):
    mcr = MCR()
    # Create a series with some variation for fitting
    idx = pd.date_range("2020-01-01", periods=20, freq="D")
    varied_data = np.sin(np.linspace(0, 2 * np.pi, 20)) + 15  # Sine wave around 15
    varied_wt = pd.Series(varied_data, index=idx)

    mcr.fit_mcr(varied_wt)
    ax = mcr.plot(varied_wt)
    assert ax is not None



import numpy as np
import pandas as pd
import pytest

from gwtf.model import Model


@pytest.fixture
def simple_wt():
    idx = pd.date_range("2020-01-01", periods=10, freq="D")
    data = np.linspace(10, 20, 10)
    return pd.Series(data, index=idx)


def test_model_init_valid(simple_wt):
    m = Model(simple_wt)
    assert isinstance(m.wt, pd.Series)
    assert m.wt.index.is_monotonic_increasing


def test_model_init_invalid():
    with pytest.raises(ValueError):
        Model([1, 2, 3])


def test_get_recharge_event_intervals(simple_wt):
    m = Model(simple_wt)
    events = m.get_recharge_event_intervals(m.wt)
    assert hasattr(events, "left")
    assert hasattr(events, "right")


def test_estimate_recharge(simple_wt):
    m = Model(simple_wt)
    recharge = m.estimate_recharge(sy=0.1)
    assert isinstance(recharge, pd.Series)
    assert (recharge >= 0).all()


def test_estimate_recharge_no_rises():
    idx = pd.date_range("2020-01-01", periods=10, freq="D")
    data = np.linspace(20, 10, 10)  # strictly decreasing
    wt = pd.Series(data, index=idx)
    m = Model(wt)
    with pytest.raises(ValueError):
        m.get_recharge_event_intervals(m.wt)


def test_plot_runs(simple_wt):
    m = Model(simple_wt)
    m.estimate_recharge(sy=0.1)
    axs = m.plot()
    assert axs is not None



import pytest

from gwtf.core import validate_data


def test_validate_data_valid(valid_series):
    # Should not raise any error
    validate_data(valid_series)


def test_validate_data_not_series(not_a_series):
    with pytest.raises(ValueError, match="Pandas Series"):
        validate_data(not_a_series)


def test_validate_data_non_datetime_index(non_datetime_index_series):
    with pytest.raises(ValueError, match="DatetimeIndex"):
        validate_data(non_datetime_index_series)


def test_validate_data_empty(empty_series):
    with pytest.raises(ValueError, match="empty"):
        validate_data(empty_series)



"""The file contains the methods to estimate the Master Recession Curve (MCR).

Raoul Collenteur, 2025
"""

import matplotlib.pyplot as plt
from pandas import DataFrame, IntervalIndex, Series
from scipy.optimize import curve_fit


class MCR:
    def __init__(self):
        """Master Recession Curve (MCR) class to fit and apply the MCR.

        Notes
        -----
        The MCR is a simple linear recession model that can be used to estimate the
        recession of the water table. The MCR is defined as:

            dh/dt = -a * h + b

        where h is the water table, a and b are the parameters of the MCR.

        The MCR can be fitted to the falling sections of the water table fluctuations.
        The falling sections are defined as the sections where the water table is
        decreasing. The MCR can be used to extrapolate the water table during rise
        events.

        """
        self.name = "MCR"
        self.nparam = 2
        self.parameters = DataFrame(
            index=["a", "b"],
            data=[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
            columns=["initial", "optimal", "stderr"],
        )
        self.fall_sections = None
        self.dhdt = None

    def get_fall_sections(self, wt: Series) -> DataFrame:
        """Get the fall sections of the water table fluctuations.

        Parameters
        ----------
        wt : pd.Series
            Time series of the water table fluctuations.

        Returns
        -------
        pd.DataFrame
            DataFrame with the fall sections of the water table fluctuations.
        """
        dt = wt.index.diff()[1:]
        dh = wt.diff()[1:]

        # Create an IntervalIndex for which the rises are computed
        dtint = IntervalIndex.from_arrays(left=dh.index - dt, right=dh.index)

        changes = DataFrame(index=dtint, data=dh.values)

        # Only keep the falls that are negative
        fall_sections = changes[changes.values < 0]

        return fall_sections

    def get_dhdt(self, wt: Series) -> Series:
        """Get the dh/dt values from the water table fluctuations.

        Parameters
        ----------
        wt : pd.Series
            Time series of the water table fluctuations.

        Returns
        -------
        pd.Series
            Series with the dh/dt values of the water table fluctuations.

        Notes
        -----
        The dh/dt values are only computed for the falling sections of the water table
        fluctuations.

        """
        # Compute the time differences in days
        dt = wt.index.diff()[1:].total_seconds() / (3600 * 24)

        # Compute the differences in water table
        dh = wt.diff()[1:]

        # Compute dh/dt
        dhdt = dh / dt
        dhdt.index = dh.index

        return dhdt[dhdt < 0]

    def estimate_dhdt(self, wt, a=None, b=None):
        """estimate_dhdt from parameters and the water table

        Parameters
        ----------
        wt : pd.Series
            Time series of the water table fluctuations.
        a : float, optional
            Parameter a of the MCR, by default None
        b : float, optional
            Parameter b of the MCR, by default None

        Returns
        -------
        pd.Series
            Series with the estimated dh/dt values of the water table fluctuations.

        Notes
        -----
        If a and b are not provided, the optimal parameters from the fit are used.

        """
        if a is None or b is None:
            a, b = self.parameters.optimal.values

        dhdt_est = -a * wt + b
        return dhdt_est

    def fit_mcr(self, wt: Series):
        """Method to fit the Master Recession Curve (MCR) to the water table data.

        Parameters
        ----------
        wt : pd.Series
            Time series of the water table fluctuations.

        Returns
        -------
        None

        Notes
        -----
        This method is not implemented yet.
        """
        # Fit the MCR to the water table data using curve_fit
        dhdt = self.get_dhdt(wt)
        self.dhdt = dhdt
        popt, pcov = curve_fit(
            f=self.estimate_dhdt, xdata=wt.loc[dhdt.index], ydata=dhdt
        )

        self.parameters.optimal = popt
        self.parameters.stderr = pcov.diagonal() ** 0.5

    def get_extrapolated(self, wt: Series, events: IntervalIndex, p=None) -> Series:
        """Internal method to get the extrapolated values using the MCR.

        Parameters
        ----------
        wt : pd.Series
            Time series of the water table fluctuations.
        events : pd.IntervalIndex
            IntervalIndex with the rise sections of the water table fluctuations.

        Returns
        -------
        pd.Series
            Series with the extrapolated values using the MCR.

        Notes
        -----
        This method is not implemented yet.
        """
        if p is None:
            a, b = self.parameters.optimal.values

        recession = self.estimate_dhdt(wt[events.left], a, b)
        return wt[events.left] + recession.values

    def plot(self, wt) -> plt.Axes:
        """Plot the fittedmaster recession curve for diagnostic checking."""
        _, ax = plt.subplots(figsize=(6, 4))

        dhdt_est = self.estimate_dhdt(wt.loc[self.dhdt.index])

        plt.plot(
            wt.loc[self.dhdt.index], -self.dhdt, marker=".", linestyle=" ", color="k"
        )
        plt.plot(wt.loc[self.dhdt.index], -dhdt_est, color="C1")

        ax.set_xlabel("Water table (m asl)")
        ax.set_ylabel("$-dh/dt$ [L/T]")

        return ax



"""Main file with the methods to estimate the recharge.

Author: Raoul A. Collenteur, July 2025

"""

import pandas as pd


def validate_data(wt):
    """Internal method to validate the data.

    Parameters
    ----------
    wt : pd.Series
        Time series of the water table fluctuations.

    Raises
    ------
    ValueError
        If the data is not a Pandas Series or if the index is not a DatetimeIndex.

    """
    if not isinstance(wt, pd.Series):
        raise ValueError("The data should be a Pandas Series.")

    if not isinstance(wt.index, pd.DatetimeIndex):
        raise ValueError("The index of the data should be a DatetimeIndex.")

    if wt.empty:
        raise ValueError("The data is empty. Please provide a non-empty Pandas Series.")



__version__ = "0.2.0"


def show_versions():
    """Show the version of gwtf and its dependencies."""
    import matplotlib as mpl
    import numpy as np
    import pandas as pd
    import scipy as sp

    print(f"gwtf: {__version__}")
    print(f"pandas: {pd.__version__}")
    print(f"numpy: {np.__version__}")
    print(f"scipy: {sp.__version__}")
    print(f"matplotlib: {mpl.__version__}")



from gwtf.version import show_versions


def test_show_versions_runs():
    # Should not raise any error
    try:
        show_versions()
    except Exception as e:
        assert False, f"show_versions raised {e}"
